"""
`EncodingDetector` class for managing chardet encoding detection.
"""
from operator import attrgetter
from typing import List

import chardet
from rich import box
from rich.padding import Padding
from rich.table import Table

from yaralyzer.config import YaralyzerConfig
from yaralyzer.encoding_detection.encoding_assessment import ENCODING, EncodingAssessment
from yaralyzer.util.logging import log

CONFIDENCE_SCORE_RANGE = range(0, 101)


class EncodingDetector:
    """
    Manager class to ease dealing with the encoding detection library `chardet`.

    Each instance of this class manages a `chardet.detect_all()` scan on a single set of bytes.

    Attributes:
        bytes (bytes): The bytes to analyze.
        bytes_len (int): The length of the bytes.
        table (Table): A rich `Table` object summarizing the chardet results.
        assessments (List[EncodingAssessment]): List of `EncodingAssessment` objects from `chardet` results.
        unique_assessments (List[EncodingAssessment]): Unique assessments by encoding, highest confidence only.
        raw_chardet_assessments (List[dict]): Raw list of dicts returned by `chardet.detect_all()`.
        force_decode_assessments (List[EncodingAssessment]): Assessments above force decode threshold.
        force_display_assessments (List[EncodingAssessment]): Assessments above force display threshold.
        has_any_idea (Optional[bool]): `True` if `chardet` had any idea what the encoding might be,
            `False` if not, `None` if `chardet` wasn't run yet.
        force_display_threshold (float): `[class variable]` Default confidence threshold for forcing display
            in decoded table.
        force_decode_threshold (float): `[class variable]` Default confidence threshold for forcing a decode attempt.
    """

    # Default value for encodings w/confidences below this will not be displayed in the decoded table
    force_display_threshold = 20.0
    # Default value for what chardet.detect() confidence % should we force a decode with an obscure encoding.
    force_decode_threshold = 50.0

    def __init__(self, _bytes: bytes) -> None:
        """
        Args:
            _bytes (bytes): The bytes to analyze with `chardet`.
        """
        self.bytes = _bytes
        self.bytes_len = len(_bytes)
        self.table = _empty_chardet_results_table()

        # Skip chardet if there's not enough bytes available
        if not self.has_enough_bytes():
            log.debug(f"{self.bytes_len} is not enough bytes to run chardet.detect()")
            self._set_empty_results()
            self.has_any_idea = None  # not false!
            return

        # Unique by encoding, ignoring language.  Ordered from highest to lowest confidence
        self.unique_assessments = []
        self.raw_chardet_assessments = chardet.detect_all(self.bytes, ignore_threshold=True)

        if len(self.raw_chardet_assessments) == 1 and self.raw_chardet_assessments[0][ENCODING] is None:
            log.info(f"chardet.detect() has no idea what the encoding is, result: {self.raw_chardet_assessments}")
            self._set_empty_results()
            self.has_any_idea = False
            return

        self.has_any_idea = True
        self.assessments = [EncodingAssessment(a) for a in self.raw_chardet_assessments]
        self._uniquify_results_and_build_table()
        self.force_decode_assessments = self.assessments_above_confidence(type(self).force_decode_threshold)
        self.force_display_assessments = self.assessments_above_confidence(type(self).force_display_threshold)

    def get_encoding_assessment(self, encoding: str) -> EncodingAssessment:
        """
        Get the `chardet` assessment for a specific encoding.

        Args:
            encoding (str): The encoding to look for.

        Returns:
            EncodingAssessment: Assessment for the given encoding if it exists, otherwise a dummy with 0 confidence.
        """
        assessment = next((r for r in self.unique_assessments if r.encoding == encoding), None)
        return assessment or EncodingAssessment.dummy_encoding_assessment(encoding)

    def has_enough_bytes(self) -> bool:
        """Return `True` if we have enough bytes to run `chardet.detect()`."""
        return self.bytes_len >= YaralyzerConfig.args.min_chardet_bytes

    def assessments_above_confidence(self, cutoff: float) -> List[EncodingAssessment]:
        """Return the assessments above the given confidence cutoff."""
        return [a for a in self.unique_assessments if a.confidence >= cutoff]

    def __rich__(self) -> Padding:
        return Padding(self.table, (0, 0, 0, 0))

    def _uniquify_results_and_build_table(self) -> None:
        """Keep the highest result per encoding, ignoring the language `chardet` has indicated."""
        already_seen_encodings = {}

        for i, result in enumerate(self.assessments):
            if result.confidence < YaralyzerConfig.args.min_chardet_table_confidence:
                continue

            self.table.add_row(f"{i + 1}", result.encoding_label, result.confidence_text)

            # self.unique_assessments retains one result per encoding possibility (the highest confidence one)
            # Some encodings are not language specific and for those we don't care about the language
            if result.encoding not in already_seen_encodings:
                self.unique_assessments.append(result)
                already_seen_encodings[result.encoding] = result
            else:
                log.debug(f"Skipping chardet result {result} (already saw {already_seen_encodings[result.encoding]})")

        self.unique_assessments.sort(key=attrgetter('confidence'), reverse=True)

    def _set_empty_results(self) -> None:
        """Set empty results for when `chardet` can't help us."""
        self.assessments = []
        self.unique_assessments = []
        self.raw_chardet_assessments = []
        self.force_decode_assessments = []
        self.force_display_assessments = []


def _empty_chardet_results_table() -> Table:
    """Returns an empty `Table` with appropriate columns for `chardet` results."""
    table = Table(
        'Rank', 'Encoding', 'Confidence',
        title='chardet.detect results',
        title_style='color(153) italic dim',
        header_style='off_white',
        style='dim',
        box=box.SIMPLE,
        show_edge=False,
        collapse_padding=True
    )

    table.columns[0].justify = 'right'
    return table



"""
Methods for computing and displaying various file hashes.
"""
import hashlib
from collections import namedtuple
from typing import Optional, Union

from rich.console import JustifyMethod
from rich.table import Column, Table

from yaralyzer.helpers.rich_text_helper import LEFT, size_text
from yaralyzer.output.rich_console import GREY

BytesInfo = namedtuple('BytesInfo', ['size', 'md5', 'sha1', 'sha256'])


def bytes_hashes_table(
    bytes_or_bytes_info: Union[bytes, BytesInfo],
    title: Optional[str] = None,
    title_justify: JustifyMethod = LEFT
) -> Table:
    """
    Build a Rich `Table` displaying the size, MD5, SHA1, and SHA256 hashes of a byte sequence.

    Args:
        bytes_or_bytes_info (Union[bytes, BytesInfo]): The `bytes` to hash, or a `BytesInfo`
            namedtuple with precomputed values.
        title (Optional[str], optional): Optional title for the table. Defaults to `None`.
        title_justify (JustifyMethod, optional): Justification for the table title. Defaults to `"LEFT"`.

    Returns:
        Table: A Rich `Table` object with the size and hash values.
    """
    if isinstance(bytes_or_bytes_info, bytes):
        bytes_info = compute_file_hashes(bytes_or_bytes_info)
    else:
        bytes_info = bytes_or_bytes_info

    table = Table(
        'Size',
        Column(size_text(bytes_info.size)),
        title=f" {title} Bytes Info" if title else None,
        title_style=GREY,
        title_justify=title_justify
    )
    table.add_row('MD5', bytes_info.md5)
    table.add_row('SHA1', bytes_info.sha1)
    table.add_row('SHA256', bytes_info.sha256)
    table.columns[1].style = 'orange3'
    table.columns[1].header_style = 'bright_cyan'
    return table


def compute_file_hashes(_bytes: bytes) -> BytesInfo:
    """
    Compute the size, MD5, SHA1, and SHA256 hashes for a given byte sequence.

    Args:
        _bytes (bytes): The `bytes` to hash.

    Returns:
        BytesInfo: `BytesInfo` namedtuple containing size, md5, sha1, and sha256 values.
    """
    return BytesInfo(
        size=len(_bytes),
        md5=hashlib.md5(_bytes).hexdigest().upper(),
        sha1=hashlib.sha1(_bytes).hexdigest().upper(),
        sha256=hashlib.sha256(_bytes).hexdigest().upper()
    )


def compute_file_hashes_for_file(file_path) -> BytesInfo:
    """
    Compute the size, MD5, SHA1, and SHA256 hashes for the contents of a file.

    Args:
        file_path (str): Path to the file to hash.

    Returns:
        BytesInfo: `BytesInfo` namedtuple containing size, md5, sha1, and sha256 values for the file contents.
    """
    with open(file_path, 'rb') as file:
        return compute_file_hashes(file.read())



"""
Methods to handle turning various objects into Rich text/table/etc representations

[Rich color names](https://rich.readthedocs.io/en/stable/appendix/colors.html)
TODO: interesting colors # row_styles[0] = 'reverse bold on color(144)' <-
"""
from sys import exit
from typing import List, Optional, Union

from rich import box
from rich.columns import Columns
from rich.panel import Panel
from rich.style import Style
from rich.text import Text

from yaralyzer.output.rich_console import BYTES_BRIGHTEST, BYTES_HIGHLIGHT, YARALYZER_THEME_DICT, console
from yaralyzer.util.logging import log

# String constants
CENTER = 'center'
FOLD = 'fold'
LEFT = 'left'
MIDDLE = 'middle'
RIGHT = 'right'

# Color meter realted constants. Make even sized buckets color coded from blue (cold) to green (go)
METER_COLORS = list(reversed([82, 85, 71, 60, 67, 30, 24, 16]))
METER_INTERVAL = (100 / float(len(METER_COLORS))) + 0.1
# Color meter extra style thresholds (these are assuming a scale of 0-100)
UNDERLINE_CONFIDENCE_THRESHOLD = 90
BOLD_CONFIDENCE_THRESHOLD = 60
DIM_COUNTRY_THRESHOLD = 25

# For the table shown by running yaralyzer_show_color_theme
MAX_THEME_COL_SIZE = 35

# Text object defaults mostly for table entries
NO_DECODING_ERRORS_MSG = Text('No', style='green4 dim')
DECODING_ERRORS_MSG = Text('Yes', style='dark_red dim')


def dim_if(txt: Union[str, Text], is_dim: bool, style: Union[str, None] = None):
    """Apply 'dim' style if 'is_dim'. 'style' overrides for Text and applies for strings."""
    txt = txt.copy() if isinstance(txt, Text) else Text(txt, style=style or '')

    if is_dim:
        txt.stylize('dim')

    return txt


def meter_style(meter_pct: float | int) -> str:
    """For coloring numbers between 0 and 100 (AKA pcts). Closer to 100 means greener, closer to 0.0 means bluer."""
    if meter_pct > 100 or meter_pct < 0:
        log.warning(f"Invalid meter_pct: {meter_pct}")

    color_number = METER_COLORS[int(meter_pct / METER_INTERVAL)]
    style = f"color({color_number})"

    if meter_pct > BOLD_CONFIDENCE_THRESHOLD:
        style += ' bold'
    if meter_pct > UNDERLINE_CONFIDENCE_THRESHOLD:
        style += ' underline'

    return style


def na_txt(style: Union[str, Style] = 'white'):
    """Standard N/A text for tables and such."""
    return Text('N/A', style=style)


def newline_join(texts: List[Text]) -> Text:
    """Join a list of Text objects with newlines between them."""
    return Text("\n").join(texts)


def prefix_with_style(_str: str, style: str, root_style: Optional[Union[Style, str]] = None) -> Text:
    """Sometimes you need a Text() object to start plain lest the underline or whatever last forever."""
    return Text('', style=root_style or 'white') + Text(_str, style)


def print_fatal_error_and_exit(error_message: str) -> None:
    """
    Print a fatal error message in a `Panel` and exit.

    Args:
        error_message (str): The error message to display.
    """
    console.line(1)
    print_header_panel(error_message, expand=False, style='bold bright_red')
    console.line(1)
    exit()


def print_header_panel(headline: str, style: str, expand: bool = True, padding: tuple | None = None) -> None:
    """
    Print a headline inside a styled Rich `Panel` to the console.

    Args:
        headline (str): The text to display as the panel's headline.
        style (str): The style to apply to the panel (e.g., color, bold, reverse).
        expand (bool, optional): Whether the panel should expand to the full console width. Defaults to `True`.
        padding (tuple, optional): Padding around the panel content (top/bottom, left/right). Defaults to `(0, 2)`.
    """
    console.print(Panel(headline, box=box.DOUBLE_EDGE, expand=expand, padding=padding or (0, 2), style=style))


def reverse_color(style: Style) -> Style:
    """Reverses the color for a given style."""
    return Style(color=style.bgcolor, bgcolor=style.color, underline=style.underline, bold=style.bold)


def show_color_theme(styles: dict) -> None:
    """Print all colors in 'styles' to screen in a grid"""
    console.print(Panel('The Yaralyzer Color Theme', style='reverse'))

    colors = [
        prefix_with_style(name[:MAX_THEME_COL_SIZE], style=str(style)).append(' ')
        for name, style in styles.items()
        if name not in ['reset', 'repr_url']
    ]

    console.print(Columns(colors, column_first=True, padding=(0, 5), equal=True))


def size_in_bytes_text(num_bytes: int) -> Text:
    return Text(f"{num_bytes:,d}", 'number').append(' bytes', style='white')


def size_text(num_bytes: int) -> Text:
    """Convert a number of bytes into (e.g.) 54,213 bytes (52 KB)."""
    kb_txt = prefix_with_style("{:,.1f}".format(num_bytes / 1024), style='bright_cyan', root_style='white')
    kb_txt.append(' kb ')
    bytes_txt = Text('(', 'white') + size_in_bytes_text(num_bytes) + Text(')')
    return kb_txt + bytes_txt


def unprintable_byte_to_text(code: str, style: str = '') -> Text:
    """Used with ASCII escape codes and the like, gives colored results like '[NBSP]'."""
    style = BYTES_HIGHLIGHT if style == BYTES_BRIGHTEST else style
    txt = Text('[', style=style)
    txt.append(code.upper(), style=f"{style} italic dim")
    txt.append(Text(']', style=style))
    return txt


def yaralyzer_show_color_theme() -> None:
    """Script method to show yaralyzer's color theme. Invocable with 'yaralyzer_show_colors'."""
    show_color_theme(YARALYZER_THEME_DICT)



"""
Helper methods to work with files.
"""
from datetime import datetime
from os import listdir, path
from pathlib import Path
from typing import List, Optional


def files_in_dir(dir: Path | str, with_extname: Optional[str] = None) -> List[str]:
    """
    Returns paths for all non dot files in `dir` (optionally filtered to only those ending in 'with_extname').

    Args:
        dir (str): Directory to list files from.
        with_extname (Optional[str], optional): If set, only return files with this extension. Defaults to None.

    Returns:
        List[str]: List of file paths.
    """
    files = [path.join(dir, path.basename(file)) for file in listdir(dir) if not file.startswith('.')]
    files = [file for file in files if not path.isdir(file)]

    if with_extname:
        return files_with_extname(files, with_extname)
    else:
        return files


def files_with_extname(files: List[str], extname: str) -> List[str]:
    """Return only files from the list that end with the given `extname`."""
    return [f for f in files if f.endswith(f".{extname}")]


def load_binary_data(file_path: Path | str) -> bytes:
    """Load and return the raw `bytes` from a file."""
    with open(file_path, 'rb') as f:
        return f.read()


def load_file(file_path: Path | str) -> str:
    """Load and return the text contents of a file."""
    with open(file_path, 'r') as f:
        return f.read()


def timestamp_for_filename() -> str:
    """Returns a string showing current time in a file name friendly format."""
    return datetime.now().strftime("%Y-%m-%dT%H.%M.%S")
